/*
 * FBShot - takes screenshots from your framebuffer
 *
 * This program is free Software, see the COPYING file
 *
 * (C) Stephan Beyer, GPLed and copylefted...
 *
 *
 * converting 8bpp to 24bpp - code by Elliot Lee <sopwith@redhat.com>
 * of course I had to change some things... I didn't use it, sorry, 
 * but it didn't really work.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <asm/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <mntent.h>

#include <sys/vt.h> /* to handle vt changing */
#include <png.h> /* PNG lib */
#include "fb.h" /* to handle framebuffer ioctls */


#define	VERSION	"0.1d"

#define	DEFAULT_FB	"/dev/fb0"

/* some conversion macros */
#define RED565(x) ((((x) >> (5+6)) & 0x1f) << 3)
#define GREEN565(x) ((((x) >> 6) & 0x3f) << 2)
#define BLUE565(x) (((x) & 0x1f) << 3)

#define ALPHA1555(x) (((x) >> (5+5+5)) & 0x1)
#define RED1555(x) ((((x) >> (5+5)) & 0x1f) << 3)
#define GREEN1555(x) ((((x) >> 5) & 0x1f) << 3)
#define BLUE1555(x) (((x) & 0x1f) << 3)

struct fb_fix_screeninfo fb_fixinfo;
struct fb_var_screeninfo fb_varinfo;
struct fb_cmap fb_cmapinfo;
int interlace=PNG_INTERLACE_ADAM7;
int gray=0; /* -1 on ; 0 off ; */
int waitbfg=0; /* wait before grabbing (for -C )... */


void FatalError(void)
 {
    printf ("An error occured! Exiting now...\n");
    exit (1);
 }


void Usage(char *binary)
 {
    printf("Usage: %s [-ghi] [-{C|c} vt] [-d dev] [-s n] filename.png\n", binary);
 }

void Help(char *binary)
 {
    printf("FBShot - makes screenshots from framebuffer, v%s\n", VERSION);
    printf("\t\tby Stephan Beyer <fbshot@s-beyer.de>\n\n");
    printf("PLEASE READ THE F...ANTASTIC MANUAL FOR MORE INFORMATIONS\n\n");
    
    Usage(binary);
    
    printf("\nPossible options:\n");
/* please keep this list alphabetical */
    printf("\t-C n  \tgrab from console n, for slower framebuffers\n");
    printf("\t-c n  \tgrab from console n\n");
    printf("\t-d dev\tuse framebuffer device dev instead of default\n");
    printf("\t-g    \tsave a grayscaled PNG\n");
    printf("\t-h    \tprint this usage information\n");
    printf("\t-i    \tturns OFF interlacing\n");
    printf("\t-s n  \tsleep n seconds before making screenshot\n");

    printf("\nSend feedback !!!\n");
 }


void chvt(int num)
 {
	int fd;

	if(!(fd = open("/dev/console", O_RDWR)))
	 {
	    FatalError();
	 }

	if (ioctl(fd, VT_ACTIVATE, num))
	 {
		perror("ioctl VT_ACTIVATE");
		FatalError();
	 }

	if (ioctl(fd, VT_WAITACTIVE, num))
	 {
		perror("ioctl VT_WAITACTIVE");
		FatalError();
	 }

	close(fd);
	if (waitbfg)
	    sleep (3);
 }


char *read_framebuffer(char *device, int vt_num, char **command, int argc)
 {
    int fd, old_vt, i;
    struct vt_stat vt_info;
    char buffer[7680000]; /* is there a more elegant solution? eg with
			    a pointer char *buffer? */

    /* first get some information about the _current_ console */
    if (vt_num!=-1)
     {
	char tmp[128];
	if(-1==(fd=open("/dev/console", O_RDONLY)))
	 {
	    fprintf (stderr, "Couldn't open /dev/console for changing to %i!\n",vt_num);
	    fprintf (stderr, "Maybe you have no read permissions or you got out of range...\n");
	    if (vt_num>=1)
	     {
		printf ("Calling su, please log in as root to go on, otherwise press CTRL-D\n");
		sprintf (tmp, "su --command=\"%s", command[0]);
		for (i=1; i<argc; i++)
		 {
		    sprintf(tmp, "%s %s", &tmp, command[i]);
		 }
		sprintf(tmp, "%s\"", &tmp);
		system (tmp);
		exit (0);
	     }
	    else
	     {
		FatalError();
	     }
	 }
	if (ioctl(fd, VT_GETSTATE, &vt_info))
	 {
	    perror ("ioctl VT_GETSTATE");
	    FatalError();
	 }
	close (fd);
	old_vt=vt_info.v_active;
     }

    /* now open framebuffer device */
    if(!(fd=open(device, O_RDONLY)))
     {
	fprintf (stderr, "Error: Couldn't open %s.\n", device);
	FatalError();
     }

    if (vt_num!=-1)
     {
	chvt(vt_num); /* go there for information */
	if (waitbfg)
	    sleep (1);
     }

    
    if (ioctl(fd, FBIOGET_FSCREENINFO, &fb_fixinfo))
     {
	perror ("ioctl FBIOGET_FSCREENINFO");
	FatalError();
     } /* get fixed info */
    if (ioctl(fd, FBIOGET_VSCREENINFO, &fb_varinfo))
     {
	perror ("ioctl FBIOGET_VSCREENINFO");
	FatalError();
     } /* get variable info */
#if 0
    if (ioctl(fd, FBIOGETCMAP, &fb_cmapinfo))
     {
	perror ("ioctl FBIOGETCMAP");
	FatalError();
     } /* get cmap */
#endif


    if (vt_num!=-1)
	chvt(vt_info.v_active); /* go back for printing */

    printf ("\007Framebuffer %s is %i bytes...\nGrabbing %i x %i... ",
	device, 
	(fb_varinfo.xres * fb_varinfo.yres * (fb_varinfo.bits_per_pixel << 3)), 
	fb_varinfo.xres, fb_varinfo.yres);

    if (vt_num!=-1)
     {
	chvt(vt_num); /* go there for reading */
	if (waitbfg)
	    sleep (3);
     }

    read(fd, buffer, (fb_varinfo.xres * fb_varinfo.yres * (fb_varinfo.bits_per_pixel << 3))); 
	/* read full framebuffer image data */

    if (vt_num!=-1)
	chvt(vt_info.v_active); /* go back for printing */

    printf ("succeeded\n");
    
    close (fd);

    return buffer;
 }


char *convert8to24(int width, int height, char *buffer)
 {
    int i;
    char nbuffer[7680000];
    
    for (i=0; i<height*width; i++)
     {
        unsigned char c=buffer[i];
        unsigned char outc[3];
        unsigned short r[255], g[255], b[255];

	outc[0]=r[c]>>8;
	outc[1]=g[c]>>8;
	outc[2]=b[c]>>8;
	outc[3]=0>>8;
	
	sprintf(nbuffer, "%s%s", &nbuffer, outc);
     }
    
    return nbuffer;
 }

char *convert1555to32(int width, int height, char *buffer)
 {
    int i;
    char nbuffer[7680000];
    
    for (i=0; i<height*width; i++)
     {
        __u16 *cb=(__u16 *) &buffer[i];
        unsigned char outc[3];
        unsigned short r[255], g[255], b[255];

	outc[0]=(char)RED1555(*cb);
	outc[1]=(char)GREEN1555(*cb);
	outc[2]=(char)BLUE1555(*cb);
	outc[3]=(char)0; 
	/* (char)ALPHA1555(*cb) may be the same, because alpha should 
	    be 0 */
	
	sprintf(nbuffer, "%s%s", &nbuffer, outc);
     }
    
    return nbuffer;
 }

char *convert565to32(int width, int height, char *buffer)
 {
    int i;
    char nbuffer[7680000];
    
    for (i=0; i<height*width; i++)
     {
        __u16 *cb=(__u16 *) &buffer[i];
        unsigned char outc[3];
        unsigned short r[255], g[255], b[255];

	outc[0]=(char)RED565(*cb);
	outc[1]=(char)GREEN565(*cb);
	outc[2]=(char)BLUE565(*cb);
	outc[3]=(char)0;
	
	sprintf(nbuffer, "%s%s", &nbuffer, outc);
     }
    
    return nbuffer;
 }


int Write_PNG(char *buffer, char *filename, int width, int height, int bits)
 {
    png_bytep row_pointers[height];

    png_structp png_ptr;
    png_infop info_ptr;
    int i;
    int bit_depth=0, color_type;

    FILE *OUTfd = fopen(filename, "wb");

    for (i=0; i<height; i++)
	row_pointers[i]=buffer+i*4*width;

    printf ("\007Now writing PNG file - Burn All Gifs! yeah :-)\n");

    if (!OUTfd)
     {
	fprintf (stderr, "Error: Couldn't fopen %s.\n", "fb.png");
	FatalError();
     }
    
    png_ptr = png_create_write_struct(
	PNG_LIBPNG_VER_STRING, 
	(png_voidp)NULL, (png_error_ptr)NULL, (png_error_ptr)NULL);
	
    if (!png_ptr)
     {
	fprintf (stderr, "Error: Couldn't create PNG write struct.\n");
	FatalError();
     }

    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
     {
	png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
	fprintf (stderr, "Error: Couldn't create PNG info struct.\n");
	FatalError();
     }
    
    png_init_io(png_ptr, OUTfd);
    
    png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);
    
    /* here please add some things, please, please, please */

    switch(bits) {
	case 8:
/*	     buffer=convert8to24(width, height, buffer); */

	    /* ... I do a break instead... */
	    break;
	case 24: /* NOTHING TESTED */
	    if (gray)
	     {
	     	bit_depth=8;
		color_type=PNG_COLOR_TYPE_GRAY;
	     } else {
		bit_depth=4;
		color_type=PNG_COLOR_TYPE_RGB;
	     }
	    png_set_invert_alpha(png_ptr);
	    png_set_bgr(png_ptr);

	    printf ("NOTE: \033[1m24 bit\033[m mode (this) has never been tested!\n\n");
	    printf ("\033[31m====================================================\033[m\n");
	    printf ("\033[1m\033[31mMAIL ME TO TELL WHETHER THIS WORKS CORRECT OR NOT!!!\033[m\n");
	    printf ("\033[31m====================================================\033[m\n\n");
	    printf ("\033[32m\033[1mmailto:fbshot@s-beyer.de\033[m\n\n");
	    break;
	case 15:
	    /* 5/6/5 */
/*	    buffer=convert565to32(width, height, buffer);*/

	    /* and then the same as 32 bit */
	    break;
	case 16:
    	    /* 1/5/5/5 */
/*	    buffer=convert1555to32(width, height, buffer);*/
	    
	    /* and then the same as 32 bit */

	    break;
	case 32:
	    if (gray)
	     {
		bit_depth=16; /* not as original as converting would be, */
		color_type=PNG_COLOR_TYPE_GRAY_ALPHA; /* but it looks good. */
	     } else {
		bit_depth=8;
		color_type=PNG_COLOR_TYPE_RGB_ALPHA;
	     }

	    png_set_invert_alpha(png_ptr);
	    png_set_bgr(png_ptr);

	    break;
     }
    
    if (bit_depth==0)
     {
	fprintf (stderr, "%d bits per pixel are not yet supported! ", bits);
	fprintf (stderr, "But you may write it...\n");
	FatalError();
     }

    png_set_IHDR(png_ptr, info_ptr, width, height, 
	bit_depth, color_type, interlace, 
	PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
    
    png_write_info(png_ptr, info_ptr);

    png_write_image(png_ptr, row_pointers);

    png_write_end(png_ptr, info_ptr);
    /* puh, done, now freeing memory... */
    png_destroy_write_struct(&png_ptr, &info_ptr);

    fclose(OUTfd);
    return 0;
 }


/********
 * MAIN *
 ********/

int main(int argc, char **argv)
 {
    char *buffer, *device = NULL, *outfile = argv[argc-1];
    int optc, vt_num=-1;

    for(;;)
     {
	optc=getopt(argc, argv, "ghiC:c:d:s:");
	if (optc==-1)
	    break;
	switch (optc) {
/* please keep this list alphabetical */ 
	case 'C':
	    waitbfg=1;
	case 'c':
	    vt_num=atoi(optarg);
	    break;
	case 'd':
	    device=optarg;
	    break;
	case 'g':
	    gray=-1;
	    break;
	case 'h':
	    Help(argv[0]);
	    return 1;
	    break;
	case 'i':
	    interlace=PNG_INTERLACE_NONE;
	    break;
	case 's':
	    sleep (atoi(optarg));
	    break;
	default:
	    Usage(argv[0]);
	}
     }
    
    if ((optind==argc) ||
	(1!=argc-optind))
     {
	Usage(argv[0]);
	return 1;
     }
    
    if (NULL==device)
     {
        device=getenv("FRAMEBUFFER");
	if (NULL==device)
	 {
	    device=DEFAULT_FB;
	 }
     }

    buffer = read_framebuffer(device, vt_num, argv, argc);

    Write_PNG(buffer, outfile, fb_varinfo.xres, fb_varinfo.yres, 
					fb_varinfo.bits_per_pixel);
    
    printf ("\007succeeded!\n");
    
    return 0;
 }
